@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2013-2015 Intel Corporation, Inc.  All rights reserved.
+ * (C) Copyright 2020 Hewlett Packard Enterprise Development LP
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -37,7 +38,8 @@
 #include <stdint.h>
 #include <rdma/rdma_cma.h>
 
-#include "fi_verbs.h"
+#include "verbs_ofi.h"
+#include "verbs_osd.h"
 
 
 #define VERBS_IB_PREFIX "IB-0x"
@@ -64,7 +66,7 @@
 		OFI_ORDER_WAW_SET | FI_ORDER_WAS | FI_ORDER_SAW | FI_ORDER_SAS )
 
 #define VERBS_INFO_NODE_2_UD_ADDR(sybsys, node, svc, ib_ud_addr)			\
-	VERBS_INFO(sybsys, "'%s:%u' resolved to <gid <interface_id=%"PRIu64		\
+	VRB_INFO(sybsys, "'%s:%u' resolved to <gid <interface_id=%"PRIu64		\
 			   ", subnet_prefix=%"PRIu64">, lid=%d, service = %u>\n",	\
 		   node, svc, be64toh((ib_ud_addr)->gid.global.interface_id),		\
 		   be64toh((ib_ud_addr)->gid.global.subnet_prefix),			\
@@ -80,7 +82,7 @@ const struct fi_domain_attr verbs_domain_attr = {
 	.control_progress	= FI_PROGRESS_AUTO,
 	.data_progress		= FI_PROGRESS_AUTO,
 	.resource_mgmt		= FI_RM_ENABLED,
-	.mr_mode		= OFI_MR_BASIC_MAP | FI_MR_LOCAL | FI_MR_BASIC,
+	.mr_mode		= OFI_MR_BASIC_MAP | FI_MR_LOCAL | OFI_MR_BASIC,
 	.mr_key_size		= sizeof_field(struct ibv_sge, lkey),
 	.cq_data_size		= sizeof_field(struct ibv_send_wr, imm_data),
 	.tx_ctx_cnt		= 1024,
@@ -107,8 +109,6 @@ const struct fi_rx_attr verbs_rx_attr = {
 	.mode			= VERBS_RX_MODE,
 	.op_flags		= FI_COMPLETION,
 	.msg_order		= VERBS_MSG_ORDER,
-	.comp_order		= FI_ORDER_STRICT | FI_ORDER_DATA,
-	.total_buffered_recv	= 0,
 };
 
 const struct fi_rx_attr verbs_dgram_rx_attr = {
@@ -116,8 +116,6 @@ const struct fi_rx_attr verbs_dgram_rx_attr = {
 	.mode			= VERBS_DGRAM_RX_MODE | VERBS_RX_MODE,
 	.op_flags		= FI_COMPLETION,
 	.msg_order		= VERBS_MSG_ORDER,
-	.comp_order		= FI_ORDER_STRICT | FI_ORDER_DATA,
-	.total_buffered_recv	= 0,
 };
 
 const struct fi_tx_attr verbs_tx_attr = {
@@ -125,7 +123,6 @@ const struct fi_tx_attr verbs_tx_attr = {
 	.mode			= 0,
 	.op_flags		= VERBS_TX_OP_FLAGS,
 	.msg_order		= VERBS_MSG_ORDER,
-	.comp_order		= FI_ORDER_STRICT,
 	.inject_size		= 0,
 	.rma_iov_limit		= 1,
 };
@@ -135,7 +132,6 @@ const struct fi_tx_attr verbs_dgram_tx_attr = {
 	.mode			= 0,
 	.op_flags		= VERBS_TX_OP_FLAGS,
 	.msg_order		= VERBS_MSG_ORDER,
-	.comp_order		= FI_ORDER_STRICT,
 	.inject_size		= 0,
 	.rma_iov_limit		= 1,
 };
@@ -158,9 +154,6 @@ const struct verbs_ep_domain verbs_dgram_domain = {
 	.protocol		= FI_PROTO_UNSPEC,
 };
 
-/* The list (not thread safe) is populated once when the provider is initialized */
-DEFINE_LIST(verbs_devs);
-
 int vrb_check_ep_attr(const struct fi_info *hints,
 			 const struct fi_info *info)
 {
@@ -182,7 +175,7 @@ int vrb_check_ep_attr(const struct fi_info *hints,
 	case FI_PROTO_IB_UD:
 		break;
 	default:
-		VERBS_INFO(FI_LOG_CORE,
+		VRB_INFO(FI_LOG_CORE,
 			   "Unsupported protocol\n");
 		return -FI_ENODATA;
 	}
@@ -236,16 +229,19 @@ static int vrb_check_hints(uint32_t version, const struct fi_info *hints,
 	uint64_t prov_mode;
 
 	if (hints->caps & ~(info->caps)) {
-		VERBS_INFO(FI_LOG_CORE, "Unsupported capabilities\n");
-		FI_INFO_CHECK(&vrb_prov, info, hints, caps, FI_TYPE_CAPS);
+		VRB_INFO(FI_LOG_CORE, "Unsupported capabilities\n");
+		OFI_INFO_CHECK(&vrb_prov, info, hints, caps, FI_TYPE_CAPS);
 		return -FI_ENODATA;
 	}
 
+	if (!ofi_valid_addr_format(info->addr_format, hints->addr_format))
+		return -FI_ENODATA;
+
 	prov_mode = ofi_mr_get_prov_mode(version, hints, info);
 
 	if ((hints->mode & prov_mode) != prov_mode) {
-		VERBS_INFO(FI_LOG_CORE, "needed mode not set\n");
-		FI_INFO_MODE(&vrb_prov, prov_mode, hints->mode);
+		VRB_INFO(FI_LOG_CORE, "needed mode not set\n");
+		OFI_INFO_MODE(&vrb_prov, prov_mode, hints->mode);
 		return -FI_ENODATA;
 	}
 
@@ -259,7 +255,7 @@ static int vrb_check_hints(uint32_t version, const struct fi_info *hints,
 	if (hints->domain_attr) {
 		if (hints->domain_attr->name &&
 		    strcasecmp(hints->domain_attr->name, info->domain_attr->name)) {
-			VERBS_INFO(FI_LOG_CORE, "skipping device %s (want %s)\n",
+			VRB_INFO(FI_LOG_CORE, "skipping device %s (want %s)\n",
 				   info->domain_attr->name, hints->domain_attr->name);
 			return -FI_ENODATA;
 		}
@@ -324,25 +320,25 @@ int vrb_set_rai(uint32_t addr_format, void *src_addr, size_t src_addrlen,
 		break;
 	case FI_SOCKADDR:
 		rai->ai_port_space = RDMA_PS_TCP;
-		if (src_addrlen) {
+		if (src_addr && src_addrlen) {
 			rai->ai_family = ((struct sockaddr *)src_addr)->sa_family;
 			rai->ai_flags |= RAI_FAMILY;
-		} else if (dest_addrlen) {
+		} else if (dest_addr && dest_addrlen) {
 			rai->ai_family = ((struct sockaddr *)dest_addr)->sa_family;
 			rai->ai_flags |= RAI_FAMILY;
 		}
 		break;
 	default:
-		VERBS_INFO(FI_LOG_FABRIC, "Unknown addr_format\n");
+		VRB_INFO(FI_LOG_FABRIC, "Unknown addr_format\n");
 	}
 
-	if (src_addrlen) {
+	if (src_addr && src_addrlen) {
 		if (!(rai->ai_src_addr = malloc(src_addrlen)))
 			return -FI_ENOMEM;
 		memcpy(rai->ai_src_addr, src_addr, src_addrlen);
 		rai->ai_src_len = src_addrlen;
 	}
-	if (dest_addrlen) {
+	if (dest_addr && dest_addrlen) {
 		if (!(rai->ai_dst_addr = malloc(dest_addrlen)))
 			return -FI_ENOMEM;
 		memcpy(rai->ai_dst_addr, dest_addr, dest_addrlen);
@@ -399,26 +395,26 @@ static int vrb_rai_to_fi(struct rdma_addrinfo *rai, struct fi_info *fi)
 
 	fi->addr_format = ofi_translate_addr_format(rai->ai_family);
 	if (fi->addr_format == FI_FORMAT_UNSPEC) {
-		VERBS_WARN(FI_LOG_FABRIC, "Unknown address format\n");
+		VRB_WARN(FI_LOG_FABRIC, "Unknown address format\n");
 		return -FI_EINVAL;
 	}
 
 	if (rai->ai_src_len) {
 		free(fi->src_addr);
- 		if (!(fi->src_addr = malloc(rai->ai_src_len)))
- 			return -FI_ENOMEM;
- 		memcpy(fi->src_addr, rai->ai_src_addr, rai->ai_src_len);
- 		fi->src_addrlen = rai->ai_src_len;
- 	}
- 	if (rai->ai_dst_len) {
+		if (!(fi->src_addr = malloc(rai->ai_src_len)))
+			return -FI_ENOMEM;
+		memcpy(fi->src_addr, rai->ai_src_addr, rai->ai_src_len);
+		fi->src_addrlen = rai->ai_src_len;
+	}
+	if (rai->ai_dst_len) {
 		free(fi->dest_addr);
 		if (!(fi->dest_addr = malloc(rai->ai_dst_len)))
 			return -FI_ENOMEM;
- 		memcpy(fi->dest_addr, rai->ai_dst_addr, rai->ai_dst_len);
- 		fi->dest_addrlen = rai->ai_dst_len;
- 	}
+		memcpy(fi->dest_addr, rai->ai_dst_addr, rai->ai_dst_len);
+		fi->dest_addrlen = rai->ai_dst_len;
+	}
 
- 	return FI_SUCCESS;
+	return FI_SUCCESS;
 }
 
 static inline int vrb_get_qp_cap(struct ibv_context *ctx,
@@ -433,13 +429,13 @@ static inline int vrb_get_qp_cap(struct ibv_context *ctx,
 
 	pd = ibv_alloc_pd(ctx);
 	if (!pd) {
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC, "ibv_alloc_pd", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "ibv_alloc_pd");
 		return -errno;
 	}
 
 	cq = ibv_create_cq(ctx, 1, NULL, NULL, 0);
 	if (!cq) {
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC, "ibv_create_cq", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "ibv_create_cq");
 		ret = -errno;
 		goto err1;
 	}
@@ -475,7 +471,7 @@ static inline int vrb_get_qp_cap(struct ibv_context *ctx,
 
 	qp = ibv_create_qp(pd, &init_attr);
 	if (!qp) {
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC, "ibv_create_qp", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "ibv_create_qp");
 		ret = -errno;
 		goto err2;
 	}
@@ -550,8 +546,7 @@ static int vrb_get_device_attrs(struct ibv_context *ctx,
 
 	ret = ibv_query_device(ctx, &device_attr);
 	if (ret) {
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC,
-				 "ibv_query_device", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "ibv_query_device");
 		return -errno;
 	}
 
@@ -600,8 +595,7 @@ static int vrb_get_device_attrs(struct ibv_context *ctx,
 	for (port_num = 1; port_num < device_attr.phys_port_cnt + 1; port_num++) {
 		ret = ibv_query_port(ctx, port_num, &port_attr);
 		if (ret) {
-			VERBS_INFO_ERRNO(FI_LOG_FABRIC,
-					 "ibv_query_port", errno);
+			VRB_WARN_ERRNO(FI_LOG_FABRIC, "ibv_query_port");
 			return -errno;
 		}
 		if (port_attr.state == IBV_PORT_ACTIVE)
@@ -609,18 +603,18 @@ static int vrb_get_device_attrs(struct ibv_context *ctx,
 	}
 
 	if (port_num == device_attr.phys_port_cnt + 1) {
-		FI_WARN(&vrb_prov, FI_LOG_FABRIC, "device %s: there are no "
+		FI_INFO(&vrb_prov, FI_LOG_FABRIC, "device %s: there are no "
 			"active ports\n", dev_name);
 		return -FI_ENODATA;
 	} else {
-		VERBS_INFO(FI_LOG_FABRIC, "device %s: first found active port "
+		VRB_INFO(FI_LOG_FABRIC, "device %s: first found active port "
 			   "is %"PRIu8"\n", dev_name, port_num);
 	}
 
 	if (info->ep_attr->type == FI_EP_DGRAM) {
 		ret = vrb_mtu_type_to_len(port_attr.active_mtu);
 		if (ret < 0) {
-			VERBS_WARN(FI_LOG_FABRIC, "device %s (port: %d) reports"
+			VRB_WARN(FI_LOG_FABRIC, "device %s (port: %d) reports"
 				   " an unrecognized MTU (%d) \n",
 				   dev_name, port_num, port_attr.active_mtu);
 			return ret;
@@ -638,7 +632,7 @@ static int vrb_get_device_attrs(struct ibv_context *ctx,
 		       device_attr.vendor_part_id);
 	if (ret < 0) {
 		info->nic->device_attr->device_id = NULL;
-		VERBS_WARN(FI_LOG_FABRIC,
+		VRB_WARN(FI_LOG_FABRIC,
 			   "Unable to allocate memory for device_attr::device_id\n");
 		return -FI_ENOMEM;
 	}
@@ -647,7 +641,7 @@ static int vrb_get_device_attrs(struct ibv_context *ctx,
 		       device_attr.vendor_id);
 	if (ret < 0) {
 		info->nic->device_attr->vendor_id = NULL;
-		VERBS_WARN(FI_LOG_FABRIC,
+		VRB_WARN(FI_LOG_FABRIC,
 			   "Unable to allocate memory for device_attr::vendor_id\n");
 		return -FI_ENOMEM;
 	}
@@ -656,14 +650,14 @@ static int vrb_get_device_attrs(struct ibv_context *ctx,
 		       device_attr.hw_ver);
 	if (ret < 0) {
 		info->nic->device_attr->device_version = NULL;
-		VERBS_WARN(FI_LOG_FABRIC,
+		VRB_WARN(FI_LOG_FABRIC,
 			   "Unable to allocate memory for device_attr::device_version\n");
 		return -FI_ENOMEM;
 	}
 
-        info->nic->device_attr->firmware = strdup(device_attr.fw_ver);
+	info->nic->device_attr->firmware = strdup(device_attr.fw_ver);
 	if (!info->nic->device_attr->firmware) {
-		VERBS_WARN(FI_LOG_FABRIC,
+		VRB_WARN(FI_LOG_FABRIC,
 			   "Unable to allocate memory for device_attr::firmware\n");
 		return -FI_ENOMEM;
 	}
@@ -677,7 +671,7 @@ static int vrb_get_device_attrs(struct ibv_context *ctx,
 	info->nic->link_attr->network_type =
 		strdup(vrb_link_layer_str(port_attr.link_layer));
 	if (!info->nic->link_attr->network_type) {
-		VERBS_WARN(FI_LOG_FABRIC,
+		VRB_WARN(FI_LOG_FABRIC,
 			   "Unable to allocate memory for link_attr::network_type\n");
 		return -FI_ENOMEM;
 	}
@@ -728,6 +722,20 @@ static int vrb_have_device(void)
 	return ret;
 }
 
+static bool vrb_hmem_supported(const char *dev_name)
+{
+	if (ofi_hmem_p2p_disabled())
+		return false;
+
+	if (vrb_gl_data.peer_mem_support && strstr(dev_name, "mlx"))
+		return true;
+
+	if (vrb_gl_data.dmabuf_support && strstr(dev_name, "mlx5"))
+		return true;
+
+	return false;
+}
+
 static int vrb_alloc_info(struct ibv_context *ctx, struct fi_info **info,
 			     const struct verbs_ep_domain *ep_dom)
 {
@@ -737,6 +745,8 @@ static int vrb_alloc_info(struct ibv_context *ctx, struct fi_info **info,
 	const char *dev_name = ibv_get_device_name(ctx->device);
 	int ret;
 
+	vrb_prof_func_start(__func__);
+
 	if ((ctx->device->transport_type != IBV_TRANSPORT_IB) &&
 	    ((ep_dom->type == FI_EP_DGRAM) ||
 	    (ep_dom->protocol == FI_PROTO_RDMA_CM_IB_XRC)))
@@ -755,17 +765,20 @@ static int vrb_alloc_info(struct ibv_context *ctx, struct fi_info **info,
 		fi->caps = VERBS_MSG_CAPS;
 		*(fi->tx_attr) = verbs_tx_attr;
 		*(fi->rx_attr) = verbs_rx_attr;
+		fi->addr_format = FI_SOCKADDR_IB;
 		break;
 	case FI_EP_DGRAM:
 		fi->caps = VERBS_DGRAM_CAPS;
 		fi->mode = VERBS_DGRAM_RX_MODE;
 		*(fi->tx_attr) = verbs_dgram_tx_attr;
 		*(fi->rx_attr) = verbs_dgram_rx_attr;
+		fi->addr_format = FI_ADDR_IB_UD;
 		fi->ep_attr->msg_prefix_size = VERBS_DGRAM_MSG_PREFIX_SIZE;
 		break;
 	default:
 		assert(0);
-		return -FI_EINVAL;
+		ret = -FI_EINVAL;
+		goto err;
 	}
 
 
@@ -791,9 +804,8 @@ static int vrb_alloc_info(struct ibv_context *ctx, struct fi_info **info,
 
 	switch (ctx->device->transport_type) {
 	case IBV_TRANSPORT_IB:
-		if (ibv_query_gid(ctx, 1, 0, &gid)) {
-			VERBS_INFO_ERRNO(FI_LOG_FABRIC,
-					 "ibv_query_gid", errno);
+		if (ibv_query_gid(ctx, 1, vrb_gl_data.gid_idx, &gid)) {
+			VRB_WARN_ERRNO(FI_LOG_FABRIC, "ibv_query_gid");
 			ret = -errno;
 			goto err;
 		}
@@ -837,7 +849,7 @@ static int vrb_alloc_info(struct ibv_context *ctx, struct fi_info **info,
 		fi->domain_attr->cq_data_size = 0;
 		break;
 	default:
-		VERBS_INFO(FI_LOG_CORE, "Unknown transport type\n");
+		VRB_INFO(FI_LOG_CORE, "Unknown transport type\n");
 		ret = -FI_ENODATA;
 		goto err;
 	}
@@ -853,14 +865,16 @@ static int vrb_alloc_info(struct ibv_context *ctx, struct fi_info **info,
 		 ctx->device->name, ep_dom->suffix);
 
 	*info = fi;
+	vrb_prof_func_end(__func__);
 
 	return 0;
 err:
 	fi_freeinfo(fi);
+	vrb_prof_func_end(__func__);
 	return ret;
 }
 
-static void verbs_devs_print(void)
+static void verbs_devs_print(struct dlist_entry *verbs_devs)
 {
 	struct verbs_dev_info *dev;
 	struct verbs_addr *addr;
@@ -869,7 +883,7 @@ static void verbs_devs_print(void)
 
 	FI_INFO(&vrb_prov, FI_LOG_FABRIC,
 		"list of verbs devices found for FI_EP_MSG:\n");
-	dlist_foreach_container(&verbs_devs, struct verbs_dev_info,
+	dlist_foreach_container(verbs_devs, struct verbs_dev_info,
 				dev, entry) {
 		FI_INFO(&vrb_prov, FI_LOG_FABRIC,
 			"#%d %s - IPoIB addresses:\n", ++i, dev->name);
@@ -1043,11 +1057,16 @@ static int vrb_get_sib(struct dlist_entry *verbs_devs)
 		return -errno;
 
 	for (int dev = 0; dev < num_devices; dev++) {
+		if (!devices[dev])
+			continue;
+
 		context = ibv_open_device(devices[dev]);
+		if (!context)
+			continue;
 
 		ret = ibv_query_device(context, &device_attr);
 		if (ret)
-			continue;
+			goto close_device;
 
 		for (int port = 1; port <= device_attr.phys_port_cnt; port++) {
 			ret = ibv_query_port(context, port, &port_attr);
@@ -1071,8 +1090,10 @@ static int vrb_get_sib(struct dlist_entry *verbs_devs)
 						continue;
 
 					dev_name = strdup(ibv_get_device_name(context->device));
-					if (!dev_name)
+					if (!dev_name) {
+						rdma_freeaddrinfo(rai);
 						return -FI_ENOMEM;
+					}
 
 					ret = verbs_devs_add(verbs_devs, dev_name, rai);
 					if (ret) {
@@ -1085,6 +1106,9 @@ static int vrb_get_sib(struct dlist_entry *verbs_devs)
 				}
 			}
 		}
+
+close_device:
+		ibv_close_device(context);
 	}
 
 	ibv_free_device_list(devices);
@@ -1104,7 +1128,7 @@ static int vrb_getifaddrs(struct dlist_entry *verbs_devs)
 
 	ret = ofi_getifaddrs(&ifaddr);
 	if (ret) {
-		VERBS_WARN(FI_LOG_FABRIC,
+		VRB_WARN(FI_LOG_FABRIC,
 			   "unable to get interface addresses\n");
 		return ret;
 	}
@@ -1113,7 +1137,7 @@ static int vrb_getifaddrs(struct dlist_entry *verbs_devs)
 	if (iface) {
 		iface_len = strlen(iface);
 		if (iface_len > IFNAMSIZ) {
-			VERBS_INFO(FI_LOG_FABRIC, "iface name: %s, too long "
+			VRB_INFO(FI_LOG_FABRIC, "iface name: %s, too long "
 				   "max: %d\n", iface, IFNAMSIZ);
 
 		}
@@ -1123,7 +1147,7 @@ static int vrb_getifaddrs(struct dlist_entry *verbs_devs)
 
 	for (ifa = ifaddr; ifa; ifa = ifa->ifa_next) {
 		if (!ifa->ifa_addr || !(ifa->ifa_flags & IFF_UP) ||
-				!strcmp(ifa->ifa_name, "lo"))
+				(ifa->ifa_flags & IFF_LOOPBACK))
 			continue;
 
 		if (iface) {
@@ -1159,7 +1183,7 @@ static int vrb_getifaddrs(struct dlist_entry *verbs_devs)
 		num_verbs_ifs++;
 	}
 
-	verbs_devs_print();
+	verbs_devs_print(verbs_devs);
 
 	freeifaddrs(ifaddr);
 	return num_verbs_ifs ? 0 : -FI_ENODATA;
@@ -1192,7 +1216,26 @@ vrb_info_add_dev_addr(struct fi_info **info, struct verbs_dev_info *dev)
 	return 0;
 }
 
-static int vrb_get_srcaddr_devs(struct fi_info **info)
+/* domain name may have a "-<xxx>" suffix */
+static inline int vrb_cmp_domain_and_dev_name(const char *domain_name,
+					      const char *dev_name)
+{
+	size_t cmp_len;
+	char *s = strrchr(domain_name, '-');
+
+	if (s)
+		cmp_len = s - domain_name;
+	else
+		cmp_len = strlen(domain_name);
+
+	if (cmp_len != strlen(dev_name))
+		return -1;
+
+	return strncmp(domain_name, dev_name, cmp_len);
+}
+
+static int vrb_get_srcaddr_devs(struct dlist_entry *verbs_devs,
+				struct fi_info **info)
 {
 	struct verbs_dev_info *dev;
 	struct fi_info *fi;
@@ -1201,12 +1244,10 @@ static int vrb_get_srcaddr_devs(struct fi_info **info)
 	for (fi = *info; fi; fi = fi->next) {
 		if (fi->ep_attr->type == FI_EP_DGRAM)
 			continue;
-		dlist_foreach_container(&verbs_devs, struct verbs_dev_info,
+		dlist_foreach_container(verbs_devs, struct verbs_dev_info,
 					dev, entry) {
-			/* strncmp because we want to process XRC fi_info as
-			 * well which have a "-xrc" suffix in domain name */
-			if (!strncmp(fi->domain_attr->name, dev->name,
-				     strlen(dev->name))) {
+			if (!vrb_cmp_domain_and_dev_name(fi->domain_attr->name,
+						         dev->name)) {
 				ret = vrb_info_add_dev_addr(&fi, dev);
 				if (ret)
 					return ret;
@@ -1256,7 +1297,8 @@ static int vrb_set_info_addrs(struct fi_info *info,
 	return FI_SUCCESS;
 }
 
-static int vrb_fill_addr(struct rdma_addrinfo *rai, struct fi_info **info,
+static int vrb_fill_addr(struct dlist_entry *verbs_devs,
+			    struct rdma_addrinfo *rai, struct fi_info **info,
 			    struct rdma_cm_id *id)
 {
 	struct sockaddr *local_addr;
@@ -1270,7 +1312,7 @@ static int vrb_fill_addr(struct rdma_addrinfo *rai, struct fi_info **info,
 		goto rai_to_fi;
 
 	if (!id->verbs)
-		return vrb_get_srcaddr_devs(info);
+		return vrb_get_srcaddr_devs(verbs_devs, info);
 
 	/* Handle the case when rdma_cm doesn't fill src address even
 	 * though it fills the destination address (presence of id->verbs
@@ -1293,11 +1335,12 @@ rai_to_fi:
 				     NULL, NULL);
 }
 
-static int vrb_device_has_ipoib_addr(const char *dev_name)
+static int vrb_device_has_ipoib_addr(struct dlist_entry *verbs_devs,
+				     const char *dev_name)
 {
 	struct verbs_dev_info *dev;
 
-	dlist_foreach_container(&verbs_devs, struct verbs_dev_info, dev, entry) {
+	dlist_foreach_container(verbs_devs, struct verbs_dev_info, dev, entry) {
 		if (!strcmp(dev_name, dev->name))
 			return 1;
 	}
@@ -1306,17 +1349,22 @@ static int vrb_device_has_ipoib_addr(const char *dev_name)
 
 #define VERBS_NUM_DOMAIN_TYPES		3
 
-int vrb_init_info(const struct fi_info **all_infos)
+static int vrb_init_info(struct dlist_entry *verbs_devs,
+			 struct fi_info **all_infos)
 {
 	struct ibv_context **ctx_list;
 	struct fi_info *fi = NULL, *tail = NULL;
 	const struct verbs_ep_domain *ep_type[VERBS_NUM_DOMAIN_TYPES];
 	int ret = 0, i, j, num_devices, dom_count = 0;
 
+	vrb_prof_func_start(__func__);
+
+	vrb_devs_free(verbs_devs);
+	fi_freeinfo(*all_infos);
 	*all_infos = NULL;
 
 	if (!vrb_have_device()) {
-		VERBS_INFO(FI_LOG_FABRIC, "no RDMA devices found\n");
+		VRB_INFO(FI_LOG_FABRIC, "no RDMA devices found\n");
 		ret = -FI_ENODATA;
 		goto done;
 	}
@@ -1331,11 +1379,13 @@ int vrb_init_info(const struct fi_info **all_infos)
 				"fi_info for XRC FI_EP_MSG endpoints\n");
 	}
 
-	vrb_getifaddrs(&verbs_devs);
+	vrb_prof_func_start("vrb_getifaddrs");
+	vrb_getifaddrs(verbs_devs);
+	vrb_prof_func_end("vrb_getifaddrs");
 	if (!vrb_gl_data.iface)
-		vrb_get_sib(&verbs_devs);
+		vrb_get_sib(verbs_devs);
 
-	if (dlist_empty(&verbs_devs))
+	if (dlist_empty(verbs_devs))
 		FI_WARN(&vrb_prov, FI_LOG_FABRIC,
 			"no valid IPoIB interfaces found, FI_EP_MSG endpoint "
 			"type would not be available\n");
@@ -1349,15 +1399,30 @@ int vrb_init_info(const struct fi_info **all_infos)
 
 	ctx_list = rdma_get_devices(&num_devices);
 	if (!num_devices) {
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC, "rdma_get_devices", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "rdma_get_devices");
 		ret = -errno;
 		goto done;
 	}
 
+	vrb_prof_func_start("alloc_info_loop");
 	for (i = 0; i < num_devices; i++) {
+		if (!ctx_list[i]) {
+			FI_INFO(&vrb_prov, FI_LOG_FABRIC,
+				"skipping device: %d, "
+				"the interface may be down, faulty or disabled\n",
+				i);
+			continue;
+		}
+
+		if (vrb_gl_data.device_name &&
+		    strncasecmp(ctx_list[i]->device->name,
+				vrb_gl_data.device_name,
+				strlen(vrb_gl_data.device_name)))
+			continue;
+
 		for (j = 0; j < dom_count; j++) {
 			if (ep_type[j]->type == FI_EP_MSG &&
-			    !vrb_device_has_ipoib_addr(ctx_list[i]->device->name)) {
+			    !vrb_device_has_ipoib_addr(verbs_devs, ctx_list[i]->device->name)) {
 				FI_INFO(&vrb_prov, FI_LOG_FABRIC,
 					"skipping device: %s for FI_EP_MSG, "
 					"it may have a filtered IPoIB interface"
@@ -1366,28 +1431,43 @@ int vrb_init_info(const struct fi_info **all_infos)
 					ctx_list[i]->device->name);
 				continue;
 			}
-			if (vrb_gl_data.device_name &&
-			    strncasecmp(ctx_list[i]->device->name,
-					vrb_gl_data.device_name,
-					strlen(vrb_gl_data.device_name)))
-				continue;
 
 			ret = vrb_alloc_info(ctx_list[i], &fi, ep_type[j]);
-			if (!ret) {
-				if (!*all_infos)
-					*all_infos = fi;
-				else
-					tail->next = fi;
+			if (ret)
+				continue;
+
+			if (!*all_infos)
+				*all_infos = fi;
+			else
+				tail->next = fi;
+			tail = fi;
+
+			/* If verbs HMEM is supported, duplicate previously
+			 * allocated fi_info and apply HMEM flags.
+			 */
+			if (vrb_hmem_supported(ctx_list[i]->device->name)) {
+				fi = fi_dupinfo(fi);
+				if (!fi)
+					continue;
+
+				fi->caps |= FI_HMEM;
+				fi->tx_attr->caps |= FI_HMEM;
+				fi->rx_attr->caps |= FI_HMEM;
+				fi->domain_attr->mr_mode |= FI_MR_HMEM;
+
+				tail->next = fi;
 				tail = fi;
 			}
 		}
 	}
+	vrb_prof_func_end("alloc_info_loop");
 
 	/* note we are possibly discarding ENOMEM */
 	ret = *all_infos ? 0 : ret;
 
 	rdma_free_devices(ctx_list);
 done:
+	vrb_prof_func_end(__func__);
 	return ret;
 }
 
@@ -1536,10 +1616,10 @@ static int vrb_del_info_not_belong_to_dev(const char *dev_name, struct fi_info *
 	*info = NULL;
 
 	while (check_info) {
-		/* Use strncmp since verbs domain names would have "-<ep_type>" suffix */
-		if (dev_name && strncmp(dev_name, check_info->domain_attr->name,
-					strlen(dev_name))) {
-			/* This branch removing `check_info` entry from the list */
+		if (dev_name &&
+		    vrb_cmp_domain_and_dev_name(check_info->domain_attr->name,
+			                        dev_name)) {
+			/* remove unmatched `check_info` entry from the list */
 			cur = check_info;
 			if (prev)
 				prev->next = check_info->next;
@@ -1584,7 +1664,7 @@ static int vrb_resolve_ib_ud_dest_addr(const char *node, const char *service,
 	if (*dest_addr) {
 		VERBS_INFO_NODE_2_UD_ADDR(FI_LOG_CORE, node, svc, *dest_addr);
 	} else {
-		VERBS_INFO(FI_LOG_CORE,
+		VRB_INFO(FI_LOG_CORE,
 			   "failed to resolve '%s:%u'.\n", node, svc);
 		return -FI_ENODATA;
 	}
@@ -1627,24 +1707,27 @@ static int vrb_handle_ib_ud_addr(const char *node, const char *service,
 	uint32_t fmt = FI_FORMAT_UNSPEC;
 	int svc = VERBS_IB_UD_NS_ANY_SERVICE, ret = FI_SUCCESS;
 
-	if (node && !ofi_str_toaddr(node, &fmt, &addr, &len) &&
-	    fmt == FI_ADDR_IB_UD) {
-		if (flags & FI_SOURCE) {
-			src_addr = addr;
-			VERBS_INFO_NODE_2_UD_ADDR(FI_LOG_CORE, node,
-						  svc, src_addr);
+	if (node && !ofi_str_toaddr(node, &fmt, &addr, &len)) {
+		if (fmt == FI_ADDR_IB_UD) {
+			if (flags & FI_SOURCE) {
+				src_addr = addr;
+				VERBS_INFO_NODE_2_UD_ADDR(FI_LOG_CORE, node,
+							  svc, src_addr);
+			} else {
+				dest_addr = addr;
+				VERBS_INFO_NODE_2_UD_ADDR(FI_LOG_CORE, node,
+							  svc, dest_addr);
+			}
+			node = NULL;
 		} else {
-			dest_addr = addr;
-			VERBS_INFO_NODE_2_UD_ADDR(FI_LOG_CORE, node,
-						  svc, dest_addr);
+			free(addr);
 		}
-		node = NULL;
 	}
 
 	if (!src_addr) {
 		src_addr = calloc(1, sizeof(*src_addr));
 		if (!src_addr) {
-			VERBS_INFO(FI_LOG_CORE,
+			VRB_INFO(FI_LOG_CORE,
 			           "failed to allocate src addr.\n");
 			ret = -FI_ENODATA;
 			goto err;
@@ -1660,7 +1743,7 @@ static int vrb_handle_ib_ud_addr(const char *node, const char *service,
 				}
 			}
 
-			VERBS_INFO(FI_LOG_CORE, "node '%s' service '%s' "
+			VRB_INFO(FI_LOG_CORE, "node '%s' service '%s' "
 				                "converted to <service=%d>\n",
 				   node, service, src_addr->service);
 		}
@@ -1687,7 +1770,8 @@ out:
 	return ret;
 }
 
-static int vrb_handle_sock_addr(const char *node, const char *service,
+static int vrb_handle_sock_addr(struct dlist_entry *verbs_devs,
+				   const char *node, const char *service,
 				   uint64_t flags, const struct fi_info *hints,
 				   struct fi_info **info)
 {
@@ -1706,24 +1790,25 @@ static int vrb_handle_sock_addr(const char *node, const char *service,
 			goto out;
 	}
 
-	ret = vrb_fill_addr(rai, info, id);
+	ret = vrb_fill_addr(verbs_devs, rai, info, id);
 out:
 	rdma_freeaddrinfo(rai);
 	if (rdma_destroy_id(id))
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC, "rdma_destroy_id", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "rdma_destroy_id");
 	return ret;
 }
 
-static int vrb_get_match_infos(uint32_t version, const char *node,
+static int vrb_get_match_infos(struct dlist_entry *verbs_devs,
+				  uint32_t version, const char *node,
 				  const char *service, uint64_t flags,
 				  const struct fi_info *hints,
-				  const struct fi_info **raw_info,
+				  const struct fi_info *raw_info,
 				  struct fi_info **info)
 {
 	int ret, ret_sock_addr = -FI_ENODATA, ret_ib_ud_addr = -FI_ENODATA;
 
 	// TODO check for AF_IB addr
-	ret = vrb_get_matching_info(version, hints, info, *raw_info,
+	ret = vrb_get_matching_info(version, hints, info, raw_info,
 				       ofi_is_wildcard_listen_addr(node, service,
 								   flags, hints));
 	if (ret)
@@ -1731,9 +1816,9 @@ static int vrb_get_match_infos(uint32_t version, const char *node,
 
 	if (!hints || !hints->ep_attr || hints->ep_attr->type == FI_EP_MSG ||
 	    hints->ep_attr->type == FI_EP_UNSPEC) {
-		ret_sock_addr = vrb_handle_sock_addr(node, service, flags, hints, info);
+		ret_sock_addr = vrb_handle_sock_addr(verbs_devs, node, service, flags, hints, info);
 		if (ret_sock_addr) {
-			VERBS_INFO(FI_LOG_FABRIC,
+			VRB_INFO(FI_LOG_FABRIC,
 				   "handling of the socket address fails - %d\n",
 				   ret_sock_addr);
 		} else {
@@ -1746,7 +1831,7 @@ static int vrb_get_match_infos(uint32_t version, const char *node,
 	    hints->ep_attr->type == FI_EP_UNSPEC) {
 		ret_ib_ud_addr = vrb_handle_ib_ud_addr(node, service, flags, info);
 		if (ret_ib_ud_addr)
-			VERBS_INFO(FI_LOG_FABRIC,
+			VRB_INFO(FI_LOG_FABRIC,
 				   "handling of the IB ID address fails - %d\n",
 				   ret_ib_ud_addr);
 	}
@@ -1754,7 +1839,7 @@ static int vrb_get_match_infos(uint32_t version, const char *node,
 	if (ret_sock_addr && ret_ib_ud_addr) {
 		/* neither the sockaddr nor the ib_ud address wasn't
 		 * handled to satisfy the selection procedure */
-		VERBS_INFO(FI_LOG_CORE, "Handling of the addresses fails, "
+		VRB_INFO(FI_LOG_CORE, "Handling of the addresses fails, "
 			   "the getting infos is unsuccessful\n");
 		fi_freeinfo(*info);
 		return -FI_ENODATA;
@@ -1792,22 +1877,105 @@ void vrb_alter_info(const struct fi_info *hints, struct fi_info *info)
 	}
 }
 
+static void vrb_filter_info_by_addr_format(struct fi_info **info, int addr_format)
+{
+	struct fi_info *cur, *prev= NULL, *next = NULL;
+	for (cur = *info; cur; cur = next) {
+		next = cur->next;
+		if (!ofi_match_addr_format(cur->addr_format, addr_format)) {
+			if (prev)
+				prev->next = cur->next;
+			else
+				*info = cur->next;
+			cur->next = NULL;
+			fi_freeinfo(cur);
+		} else {
+			prev = cur;
+		}
+	}
+}
+
+void vrb_devs_free(struct dlist_entry *verbs_devs)
+{
+	struct verbs_dev_info *dev;
+	struct verbs_addr *addr;
+
+	while (!dlist_empty(verbs_devs)) {
+		dlist_pop_front(verbs_devs, struct verbs_dev_info, dev, entry);
+		while (!dlist_empty(&dev->addrs)) {
+			dlist_pop_front(&dev->addrs, struct verbs_addr, addr, entry);
+			rdma_freeaddrinfo(addr->rai);
+			free(addr);
+		}
+		free(dev->name);
+		free(dev);
+	}
+}
+
 int vrb_getinfo(uint32_t version, const char *node, const char *service,
 		   uint64_t flags, const struct fi_info *hints,
 		   struct fi_info **info)
 {
+	static bool init_done = false;
+	struct dlist_entry tmp_devs;
+	struct fi_info *tmp_info = NULL;
 	int ret;
 
-	ret = vrb_get_match_infos(version, node, service,
+	vrb_prof_func_start(__func__);
+	ofi_mutex_lock(&vrb_init_mutex);
+	if (!init_done) {
+		ret = vrb_init();
+		if (ret) {
+			ofi_mutex_unlock(&vrb_init_mutex);
+			goto out;
+		}
+		init_done = true;
+
+		ofi_mutex_lock(&vrb_info_mutex);
+		ret = vrb_init_info(&vrb_devs, &vrb_util_prov.info);
+		ofi_mutex_unlock(&vrb_info_mutex);
+		if (ret) {
+			ofi_mutex_unlock(&vrb_init_mutex);
+			goto out;
+		}
+		flags &= ~FI_RESCAN; /* No rescan needed */
+	}
+	ofi_mutex_unlock(&vrb_init_mutex);
+
+	if (flags & FI_RESCAN) {
+		dlist_init(&tmp_devs);
+		ret = vrb_init_info(&tmp_devs, &tmp_info);
+		if (ret) {
+			vrb_devs_free(&tmp_devs);
+			fi_freeinfo(tmp_info);
+			goto out;
+		}
+	}
+
+	ofi_mutex_lock(&vrb_info_mutex);
+	if (flags & FI_RESCAN) {
+		/* Swap the lists while holding the lock */
+		vrb_devs_free(&vrb_devs);
+		dlist_splice_tail(&vrb_devs, &tmp_devs);
+
+		fi_freeinfo(vrb_util_prov.info);
+		vrb_util_prov.info = tmp_info;
+	}
+	ret = vrb_get_match_infos(&vrb_devs, version, node, service,
 				     flags, hints,
-				     &vrb_util_prov.info, info);
+				     vrb_util_prov.info, info);
+	ofi_mutex_unlock(&vrb_info_mutex);
 	if (ret)
 		goto out;
 
 	ofi_alter_info(*info, hints, version);
 
 	vrb_alter_info(hints, *info);
+
+	if (hints)
+		vrb_filter_info_by_addr_format(info, hints->addr_format);
 out:
+	vrb_prof_func_end(__func__);
 	if (!ret || ret == -FI_ENOMEM || ret == -FI_ENODEV)
 		return ret;
 	else
